let
   shutdownManager = require('video-utils/shutdownManager'),
   logger = require('./../logger/logger'),
   copyFrom = require('pg-copy-streams').from,
   pgClient = require('./pgClient');

const CLOSE_TIMEOUT = 5000;

/**
 * Класс предназначен для записи в таблицу подготовленных данных методом COPY
 */
class DBStreamer {
   constructor(query, queryName) {
      //текст запроса
      this.query = query;
      //имя запроса для explain
      this.queryName = queryName;
      //Сюда складываем все пришедшие для записи сообщения
      this.archive = [];
      //доступность стрима для записи
      this.streamAvailable = false;
      //тут храним таймаут на закрытие транзакции
      this.closeTimeout = null;
      //состояние коннекта до pg
      this.pgConnectionState = DBStreamer.PG_CONNECTION_CLOSED;

      shutdownManager.addHandler(() => {
         clearTimeout(this.closeTimeout);
         return this._writeArchive();
      });
   }

   /**
    * Единственный метод наружу, служит для записи строки данных в определенную таблицу через COPY
    * @param {String} data - данные для записи, строки через /n,  значения в строках разделенны /t
    */
   copy(data) {
      this.archive.push(data);

      //если нет активного стрима и в данный момент не пересоздается подключение к базе данных
      if (!this.stream && this.pgConnectionState !== DBStreamer.PG_CONNECTION_CONNECTING) {
         Promise.resolve(true)
            .then(() => {
               switch (this.pgConnectionState) {
                  //если соединение не установлено, установим
                  case DBStreamer.PG_CONNECTION_CLOSED:
                     return this._getClient();
                  case DBStreamer.PG_CONNECTION_CONNECTED:
                     return true;
               }
            })
            .then(() => {
               //соединение установлено, откроем стрим
               this._openStream();
            });
      }
   }

   /**
    * Метод для получения нового коннекта до базы данных
    * @private
    */
   _getClient() {
      logger.log('DBStreamer', 'getting pgClient ' + this.queryName);
      this.pgConnectionState = DBStreamer.PG_CONNECTION_CONNECTING;
      return pgClient.getClient()
         .then(pgClient => {
            logger.log('DBStreamer', 'pgClient ready ' + this.queryName);
            this.pgClient = pgClient;
            this.pgConnectionState = DBStreamer.PG_CONNECTION_CONNECTED;

            this.pgClient.on('end',  () => {
               this.pgConnectionState = DBStreamer.PG_CONNECTION_CLOSED;
            });

            //задаем имя приложения - на explain будет отображаться как имя запроса
            pgClient.query("SET application_name = '" + this.queryName + "';");
         });
   }

   /**
    * Метод для записи очереди, скопившейся пока переоткрывается стрим
    * @private
    */
   _writeArchive() {
      return new Promise((resolve, reject) => {
         logger.log('DBStreamer', 'writing arhive ' + this.queryName);
         var writeElem = () => {
            if(this.archive.length > 0 && this.streamAvailable) {
               this.stream.write(this.archive.shift(), 'utf8', (err) => {
                  if(!err) {
                     writeElem();
                  } else {
                     logger.log('DBStreamer', 'write error ' + this.queryName);
                     reject();
                  }
               });
            } else {
               resolve();
            }
         };
         writeElem();
      })
   }

   /**
    * Метод для открытия стрима
    */
   _openStream() {
      logger.log('DBStreamer', 'opening stream ' + this.queryName);
      this.stream = this.pgClient.query(copyFrom(this.query));
      this.streamAvailable = true;

      this.stream.on('error', (err) => {
         this.streamAvailable = false;
         logger.error('DBStreamer', 'stream for ' + this.queryName + ' close irirregulary \n' + (err.message || err));
         //на ошибку стрима закроем его и переоткроем соединение до базы
         this.stream.end();
         this.stream.removeAllListeners();
         this.stream = null;

         logger.log('DBStreamer', 'reopening pgClient ' + this.queryName);
         this.pgClient.end(() => {
            this._getClient()
               .then(() => {
                  logger.log('DBStreamer', 'pgClient reopened ' + this.queryName);
                  this._onEnd();
               });
         });
      });

      this.stream.on('end', () => {
         logger.log('DBStreamer', 'stream closed ' + this.queryName);

         this.stream.removeAllListeners();
         this.stream = null;

         this._onEnd();
      });
      this._createCloseTimeout();
      this._writeArchive();
   }

   /**
    * Метод, переоткрывающий стрим, если предыдущий по какой - то причине был закрыт
    * @private
    */
   _onEnd() {
      clearTimeout(this.closeTimeout);
      if (this.archive.length) {
         this._openStream();
      }
   }

   /**
    * Метод, заводящий таймаут на переоткрытие стрима
    * @private
    */
   _createCloseTimeout() {
      this.closeTimeout = setTimeout(() => {
         logger.log('DBStreamer', 'closing stream by timeout  ' + this.queryName + '');
         this.streamAvailable = false;
         this.stream && this.stream.end();
      }, CLOSE_TIMEOUT)
   }
}

DBStreamer.PG_CONNECTION_CLOSED     = 0;
DBStreamer.PG_CONNECTION_CONNECTING = 1;
DBStreamer.PG_CONNECTION_CONNECTED  = 2;

module.exports = DBStreamer;