var
      request = require('request'),
      fs      = require('fs'),
      uuid    = require('node-uuid'),
      Q       = require('q');


var MAX_PART_SIZE = 1024 * 1024 * 80; //80Мб
var QUANTUM       = 1024 * 256;       //256Кб

function leadZeroFixedLength(str, l) {
   return ('000000000000000' + str).slice(-l);
}

/**
 * Генерим фейковый sid
 * @param {Object} authorInfo - информация о авторе, от имени которого заливается файл
 * @param {String} authorInfo.client - ид клиента
 * @param {String} authorInfo.face   - ид лица
 * @returns {string}
 */
function generateFakeSid (authorInfo){
   return [
      leadZeroFixedLength((authorInfo.client).toString(16), 8),
      leadZeroFixedLength((authorInfo.face).toString(16), 8),
      uuid.v4().replace(/-/g, '').slice(-4),
      uuid.v4().replace(/-/g, '').slice(-16)
   ].join('-');
}

/**
 * Загружает/обновляет файл целиком
 * @param file
 * @param params
 * @private
 */
function _uploadAll(file, params){
   var
         url = params.folderId ? params.url + params.folderId : params.url + params.fileId,
         method = params.folderId ? 'POST' : 'PUT';
   return Q.nfcall(request, {
      method: method,
      url: url,
      headers: Object.assign({}, params.defaultHeaders, {
         'Content-Length' : file.length,
         'Content-Type'   : params.contentType,
         'Content-Disposition' : "attachment;filename*=UTF-8''" + encodeURIComponent(params.fileName)
      }),
      body: file
   })
         .then(function(result){
            var res = result[0];
            if (res.statusCode && res.statusCode >= 200 && res.statusCode <= 204) {
               try{
                  return JSON.parse(res.body);
               } catch(e) {
                  throw new Error('sbis-disk result parse error, body:' + res.body);
               }
            } else {
               throw new Error("Upload failed, statusCode: " + res.statusCode);
            }
         })
}

/**
 * Загружает файл по частям
 * @param file
 * @param params
 * @private
 */
function _uploadParts(file, params){
   var href = params.folderId ? params.url + params.folderId : params.url + params.fileId;
   return Q.nfcall(request, {
      method: 'POST',
      json: true,
      url: params.url + 'upload/',
      headers: Object.assign({}, params.defaultHeaders, {
         'X-Upload-Content-Type': params.contentType,
         'X-Upload-Content-Length': file.length,
         'Content-Disposition' : "attachment;filename*=UTF-8''" + encodeURIComponent(params.fileName)
      }),
      body: {
         href: href
      }
   })
         .then(function(result){
            var res = result[0];
            if (res.statusCode && res.statusCode == 200) {
               return result[1].href;
            } else {
               throw new Error("Init of multipart uploading failed, statusCode: " + res.statusCode);
            }
         })
         .then(function(href){
            var
                  fileSize     = file.length,
                  pathsCount   = Math.ceil(fileSize / MAX_PART_SIZE),
                  request_size = Math.floor(fileSize / pathsCount),
                  start, end, path;

            //Размер части должен быть кратен 256 Кб
            request_size = request_size - request_size % QUANTUM;

            var globalPromise = Q.resolve(true);

            //Подготавливаем запросы на отправку частей файла
            for (var i = 0; i < pathsCount; i++) {
               start = i * request_size;
               end = i == (pathsCount - 1) ? fileSize : (i + 1) * request_size;

               path = file.slice(start, end);

               //Выстраиваем запросы их в цепочку
               globalPromise = (function(rD){
                  return globalPromise
                        .then(function(){
                           return Q.nfcall(request, rD);
                        })
                        .then(function(result){
                           var res = result[0];
                           if (res.statusCode && res.statusCode == 200) {
                              return res.statusCode;
                           } else {
                              throw new Error("Path was not uploaded, statusCode: " + res.statusCode);
                           }
                        })
               })({
                  method: 'PUT',
                  url: href + '&part=new',
                  headers: Object.assign({}, params.defaultHeaders, {
                     'Content-Length': path.length
                  }),
                  body: path
               });
            }

            //если все части загружены, то прокидываем урл для завершающего запроса
            return globalPromise
                  .then(function(){
                     return href;
                  });
         })
         .then(function(href){
            return Q.nfcall(request, {
               method: params.folderId ? 'POST' : 'PUT',
               url: href,
               headers: params.defaultHeaders
            })
                  .then(function(result){
                     var res = result[0];
                     if (res.statusCode && res.statusCode == 200) {
                        try{
                           return JSON.parse(res.body);
                        } catch(e) {
                           throw new Error('sbis-disk result parse error, body:' + res.body);
                        }
                     } else {
                        throw new Error("Multipart uploading failed, statusCode: " + res.statusCode);
                     }
                  })
         })
}

/**
 * Загружает файл на СБИС.Диск
 * @param {Object} params
 * @param {String} params.filePath    - расположение файла на диске
 * @param {String} params.contentType - MIME TYPE файла
 * @param {String} params.url         - адрес сервиса СБИС.Диск
 * @param {String} params.folderId    - иднтификатор папки в СБИС.Диске
 * @param {String} params.fileId      - идентификатор файла, который нужно обновить в виде fileId_versionId
 * @param {String} params.fileName    - имя файла
 * @param {Object} params.authorInfo  - информация о авторе, от имени которого заливается файл
 * @param {String} params.authorInfo.client - ид клиента
 * @param {String} params.authorInfo.face   - ид лица
 * @param {String} params.authorInfo.person - ид сервиса профилей
 */
function put(params){
   return Q.nfcall(fs.readFile, params.filePath)
         .then(function(file){
            var sid;
            if (params.authorInfo) {
               if (params.authorInfo.sid) {
                  sid = params.authorInfo.sid;
               } else {
                  sid = generateFakeSid(params.authorInfo)
               }
               params.defaultHeaders = {
                  'X-SBISSessionID': sid,
                  'X-OBJECT-FILENAME': encodeURIComponent(params.fileName),
                  'X-Spid': params.authorInfo.person
               };

               if (file.length < MAX_PART_SIZE) {
                  return _uploadAll(file, params);
               }
               else {
                  return _uploadParts(file, params);
               }
            }
         })
}

module.exports = {
   put: put
};